/*
    Kiwi
    Copyright (c) 1999-2004,2005 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <assert.h>
#include <limits.h>
#include <string.h>

#include "counters.h"
#include "hash.h"
#include "log.h"
#include "position.h"
#include "score.h"

HashTable::HashTable( unsigned n )
{
    assert( (n & (n-1)) == 0 );     // Make sure size is a power of two

    size = n;
    mask = n-1;
    mask &= ~3; // Store four positions per entry, so there is one fourth of all entries
    search_id = 0;
    table = new Entry[ size ];
}

HashTable::~HashTable()
{
    delete [] table;
}

void HashTable::reset()
{
    memset( table, 0, sizeof(Entry)*size );
}

HashTable::Entry * HashTable::probe( const Position & pos ) const
{
    Counters::hashProbes++;

    Entry *     entry   = table + (xhash(pos) & mask);

    if( entry->code == pos.hashCode ) {
        return entry;
    }

    entry++; // Probe the second slot
    if( entry->code == pos.hashCode ) {
        return entry;
    }

    entry++; // Probe the third slot
    if( entry->code == pos.hashCode ) {
        return entry;
    }

    entry++; // Probe the fourth slot
    if( entry->code == pos.hashCode ) {
        return entry;
    }

    Counters::hashProbesFailed++;

    return 0;
}

void HashTable::store( const Position & pos, const Move & move, int value, unsigned flags, int depth )
{
    Counters::hashStores++;

    Entry *     entry   = table + (xhash(pos) & mask);

    /*
        Replacement rules (in order of enforcement):
        1) entries with the same hash code;
        2) entries from a different search;
        3) entries with a smaller search depth.
    */
    int index = 0;

    for( int i=0; i<4; i++ ) {
        if( entry[i].code == pos.hashCode ) {
            // Position is already in the table

            /*
            if( entry[i].getSearchId() == search_id && entry->getDepth() > depth ) {
                return;
            }
            */
            break;
        }

        if( entry[i].getSearchId() != search_id ) {
            // Entry from different search: replace if current choice is from this search or less deep
            if( entry[index].getSearchId() == search_id || entry[index].getDepth() > entry[i].getDepth() ) {
                index = i;
            }
        }
        else {
            // Entry from same search: replace only if current choice is from this search and less deep
            if( entry[index].getSearchId() == search_id && entry[index].getDepth() > entry[i].getDepth() ) {
                index = i;
            }
        }
    }

    entry[index].packData1( move, flags, search_id );
    entry[index].packData2( value, depth );
    entry[index].code = pos.hashCode;
}

void HashTable::clean( const Position & pos )
{
    Entry *     entry   = table + (xhash(pos) & mask);

    entry[0].reset();
    entry[1].reset();
    entry[2].reset();
    entry[3].reset();
}

/* Old code follows... */
#if 0

/*
    Hashing scheme: 
    - NEW always replaces existing entry;
    - DEEP only replaces an entry if the new one was searched deeper;
    - TWODEEP is same as DEEP but existing entry is kept as a secondary entry and can be probed;
*/
#define SCHEME  TWODEEP

HashTable::HashTable( unsigned n )
{
    assert( (n & (n-1)) == 0 );     // Make sure size is a power of two

    size = n;
    mask = n-1;
    search_id = 0;
    table = new Entry[ size ];

    mask &= ~1; // Store two positions per entry, so there actually are half of all entries
}

HashTable::~HashTable()
{
    delete [] table;
}

void HashTable::reset()
{
    memset( table, 0, sizeof(Entry)*size );
}

HashTable::Entry * HashTable::probe( const Position & pos ) const
{
    Counters::hashProbes++;

    Entry *     entry   = table + (xhash(pos) & mask);

    if( entry->isValid() ) {
#if (SCHEME==TWODEEP)
        if( entry->code == pos.hashCode ) return entry;
        entry++; // Probe the second slot
#endif
        if( entry->code == pos.hashCode ) {
            return entry;
        }
    }

    Counters::hashProbesFailed++;

    return 0;
}

void HashTable::store( const Position & pos, const Move & move, int value, unsigned flags, int depth )
{
    Counters::hashStores++;

    Entry *     entry   = table + (xhash(pos) & mask);

#if (SCHEME==TWODEEP)
    if( entry->isValid() ) {
        // Check whether this position must be stored into the first or second slot
        if( (search_id != entry->getSearchId()) || (depth >= entry->getDepth()) ) {
            // First position has to be replaced, so move it into the second slot
            if( entry->code != pos.hashCode ) {
                Entry * entry2  = entry+1;

                entry2->data1 = entry->data1;
                entry2->data2 = entry->data2;
                entry2->code  = entry->code;
            }
        }
        else {
            // Store directly into second slot
            entry++;
        }
    }
#elif (SCHEME==DEEP)
    // Do not overwrite a deeper entry
    if( entry->isValid() && (entry->getDepth() >= depth) ) {
        return;
    }
#endif

    entry->packData1( move, flags, search_id );
    entry->packData2( value, depth );
    entry->code = pos.hashCode;
}

void HashTable::clean( const Position & pos )
{
    Entry *     entry   = table + (xhash(pos) & mask);

#if (SCHEME==TWODEEP)
    entry[1].reset();
#endif
    entry->reset();
}

#endif // OLD_CODE
