/*
    Kiwi
    Copyright (c) 1999-2004,2005 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    Silman's rules

    Minor pieces:
    1) Knights and bishops have the same point-value
    2) Bishops are best in open positions (no center pawns)
    3) Bishops are very strong in endgames where both sides have
       passed pawns
    4) A "bad" bishop is situated on the same color of its center
       pawns. Then you want to:
       - trade it;
       - get the pawns off the bishop color;
       - get the bishop outside the pawn chain (make it active)
    5) Two bishops are strong because they can control both colored
       diagonals
    6) Knights love locked positions with blocked pawns
    7) Knights are better placed in the center
    8) Knights need to have support points. As placement goes:
       - 1st rank: not effective, defense only;
       - 2nd rank: defensive, inferior to bishop;
       - 3rd rank: good defense, ready to jump elsewhere;
       - 4th rank: good, fully equal to bishop;
       - 5th rank: powerful attacking unit, usually stronger than bishop;
       - 6th rank: most powerful, very strong attacker;
       - 7th rank: not as good as 6th;
       - 8th rank: not as good as 6th.
    9) Knights are better than bishops in endgame if all pawns are on
       one side of the board. On the other side bishops are better in a
       pawn race situation where there are pawns on both sides of the
       board.
    10)To help knights it's best to create advanced support points.
       This is expecially true when fighting a knight vs. bishop battle.

    Center:
    1) A full pawn center gives territory and control over central squares
    2) A full pawn center must be protected and made indestructible
    3) Don't advance the center too early
    4) If the opponent has a full pawn center, attack it
    5) If the center pawns get traded then open files exist that help
       rooks get into play
    6) If the center becomes locked the play switches to the wings.
    7) If the center is closed, the pawn chain is used to select which wing
       to play (the one where there is more space)
    .. Push the pawn next to your most advanced pawn.
    8) With a wide open center it's better to attack with pieces, with a
       closed center pawns are usually better (to grab space and open files)

    Space:
    1) When you have more space it's better to avoid exchanges
    2) If you have less space look for exchanges as the rest of your pieces
       will have more room

    Pawn structure:
    1) Doubled pawns are less flexible and more vulnerable to attack
    2) Doubled pawns allows for extra open files and increased square control
    3) An isolated pawn is very vulnerable if it stands on an open file
    4) An isolated pawn may lead to the creation and use of an half-open file
    5) An isolated pawn is better attacked by controlling the weak square
       in front of it, trading minor pieces and attacking with the queen and
       one or two rooks (plus eventually a pawn)
    6) An isolated d4 (or d5 for black) pawn gives space on open files for the
       rooks, so it's not weak. Play should be dynamic
    7) A backward pawn is only weak if it's sitting on an open file and unable
       to advance. It's best attacked by controlling the square in front of it
       first.
    8) A backward pawn helps a more advanced pawn. It is not considered bad if
       the square in front of it is well defended
    9) Hanging pawns must move and advance otherwise they become weak
    10)The traditional hanging pawns on c5 and d5 (or c4 and d4 for white) are
       quite good
    11)A protected passed pawn is not always an advantage. If the pawn is blocked
       by an enemy piece (especially a knight) then it is much less powerful.
    12)A passed pawn is strong if it's possible to play elsewhere (gets an endgame
       advantage) or if the square in front of it is cleared for advance.
*/
#include "board.h"
#include "log.h"
#include "score.h"

/*
    Basic piece values. These can be redefined at will, however most
    other values and tables assume that the value of a pawn is 100.

*/
int Score::Pawn         = 100;
int Score::Knight       = 325;
int Score::Bishop       = 330;
int Score::Rook         = 500;
int Score::Queen        = 975;
int Score::King         = 0;

int Score::Piece[16];
int Score::PieceAbs[16];

const char * Score::ByPiece_Opening[16];
const char * Score::ByPiece_Endgame[16];

/*
    Adjustment is a ratio expressed in 1/1000th units and is added or subtracted from
    the corresponding piece/square values. For example, a +100 adjustment causes
    all values to be incremented by 10 percent. Use 0 for no adjustment.
*/
const int   Knight_Adjustment = 0;
const int   Bishop_Adjustment = 0;
const int   Rook_Adjustment = 0;
const int   Queen_Adjustment = 0;

static int adjustScore( int score, int adjustment )
{
    return score + (score * adjustment) / 1000;
}

/*
    Piece/square tables must be visualized like playing for white,
    even if in practice they refer to black because the A1 square
    is defined as zero, which happens to be in the top left corner
    when declaring arrays.

    This is especially true for tables that explicitly state "Black"
    in their name, such as for example BlackKnightOutpost. Again,
    the table is for black but it is much easier to visualize it
    from the white side, with the first rank in the bottom.
*/

char Score::BlackKnight_Opening[64] =  {
      -10, -5, -5, -5, -5, -5, -5,-10,
       -5,  0,  0,  0,  0,  0,  0,-10,
        0,  5,  7, 10, 10,  7,  5,  0,
       -2,  3,  3,  7,  7,  3,  3, -2,
       -5,  0,  3,  5,  5,  3,  0, -5,
       -5,  0,  5,  0,  0,  5,  0, -5,
      -10, -5,  0,  0,  0,  0, -5,-10,
      -15,-10, -5, -5, -5, -5,-10,-15
};

char Score::BlackKnight_Endgame[64] = {
      -15,-10, -5, -5, -5, -5,-10,-15,
      -10, -5,  0,  0,  0,  0, -5,-10,
       -5,  0,  0,  3,  3,  0,  0, -5,
       -5,  0,  3,  8,  8,  3,  0, -5,
       -5,  0,  3,  8,  8,  3,  0, -5,
       -5,  0,  0,  3,  3,  0,  0, -5,
      -10, -5,  0,  0,  0,  0, -5,-10,
      -15,-10, -5, -5, -5, -5,-10,-15
};

char Score::BlackBishop_Opening[64] =  {
      -12,-10, -5, -5, -5, -5,-10,-12,
      -10, -7,  0,  0,  0,  0, -7,-10,
      -10,  0,  0,  0,  0,  0,  0,-10,
       -5,  0,  0,  0,  0,  0,  0, -5,
       -5,  0,  0,  2,  2,  0,  0, -5,
       -5,  0,  2,  0,  0,  2,  0, -5,
       -8,  2,  0,  0,  0,  0,  2, -8,
      -10, -8, -5, -5, -5, -5, -8,-10
};

char Score::BlackBishop_Endgame[64] = {
       -8, -5, -3, -3, -3, -3, -5, -8,
       -5, -3,  0,  0,  0,  0, -3, -5,
       -3,  0,  0,  0,  0,  0,  0, -3,
       -3,  0,  0,  3,  3,  0,  0, -3,
       -3,  0,  0,  3,  3,  0,  0, -3,
       -3,  0,  0,  0,  0,  0,  0, -3,
       -5, -3,  0,  0,  0,  0, -3, -5,
       -8, -5, -3, -3, -3, -3, -5, -8,
};

char Score::BlackRook_Opening[64] =  {
       -9,  0,  0,  0,  0,  0,  0, -9,
       -9,  0,  0,  0,  0,  0,  0, -9,
       -7,  0,  0,  0,  0,  0,  0, -7,
       -7,  0,  0,  0,  0,  0,  0, -7,
       -5,  0,  0,  0,  0,  0,  0, -5,
       -5,  0,  0,  3,  3,  0,  0, -5,
       -3,  0,  3,  6,  6,  3,  0, -3,
       -3,  0,  3,  6,  6,  3,  0, -3,
};

char Score::BlackRook_Endgame[64] = {
       -3,  0,  0,  3,  3,  0,  0, -3,
        0,  0,  0,  3,  3,  0,  0,  0,
        0,  0,  0,  3,  3,  0,  0,  0,
        3,  3,  3,  5,  5,  3,  3,  3,
        3,  3,  3,  5,  5,  3,  3,  3,
        0,  0,  0,  3,  3,  0,  0,  0,
        0,  0,  0,  3,  3,  0,  0,  0,
       -3,  0,  0,  3,  3,  0,  0, -3
};

char Score::BlackQueen_Opening[64] =  {
      -15,-10,-10,-10,-10,-10,-10,-15,
      -15, -5,  0,  0,  0,  0, -5,-15,
      -10, -5,  0,  0,  0,  0, -5,-10,
      -10, -2,  0,  0,  0,  0, -2,-10,
      -10, -2,  0,  0,  0,  0, -2,-10,
      -10,  0,  0,  0,  0,  0,  0,-10,
      -10,  0,  3,  3,  3,  3,  0,-10,
      -15,-10,  0, -5, -5,  0,-10,-15
};

char Score::BlackQueen_Endgame[64] = {
       -8, -5, -5, -5, -5, -5, -5, -8,
       -5,  0,  0,  0,  0,  0,  0, -5,
       -5,  0,  3,  5,  5,  3,  0, -5,
       -5,  0,  5,  8,  8,  5,  0, -5,
       -5,  0,  5,  8,  8,  5,  0, -5,
       -5,  0,  3,  5,  5,  3,  0, -5,
       -5,  0,  0,  0,  0,  0,  0, -5,
       -8, -5, -5, -5, -5, -5, -5, -8
};

char Score::BlackKing_Opening[64] =  {
      -30,-30,-30,-30,-30,-30,-30,-30,
      -30,-30,-30,-30,-30,-30,-30,-30,
      -20,-20,-20,-20,-20,-20,-20,-20,
      -20,-20,-20,-20,-20,-20,-20,-20,
      -20,-20,-20,-20,-20,-20,-20,-20,
      -20,-20,-20,-20,-20,-20,-20,-20,
      -10,-10,-10,-15,-15,-10,-10,-10,
        0,  0,  5,-10,-10, -5,  5,  0
};

char Score::BlackKing_Endgame[64] = {
       -8, -5, -3, -3, -3, -3, -5, -8,
       -5, -3,  0,  0,  0,  0, -3, -5,
       -3,  0,  3,  4,  4,  3,  0, -3,
       -3,  0,  4,  8,  8,  4,  0, -3,
       -3,  0,  4,  8,  8,  4,  0, -3,
       -3,  0,  3,  4,  4,  3,  0, -3,
       -5, -3,  0,  0,  0,  0, -3, -5,
       -8, -5, -3, -3, -3, -3, -5, -8,
};

// Knight outpost
int Score::BlackKnightBlocker[64] = {
    0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0, 15, 15,  0,  0,  0,
    0,  5, 10, 15, 15, 10,  5,  0,
    0, 10, 15, 20, 20, 15, 10,  0,
   10, 20, 20, 20, 20, 20, 20, 10,
   20, 30, 30, 30, 30, 30, 30, 20
};

int Score::BlackKnightOutpost[64] = {
    0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  5,  5,  0,  0,  0,
    0,  0,  5, 10, 10,  5,  0,  0,
    0,  0,  5, 10, 10,  5,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0
};

char Score::WhiteKnight_Opening[64];
char Score::WhiteKnight_Endgame[64];
char Score::WhiteBishop_Opening[64];
char Score::WhiteBishop_Endgame[64];
char Score::WhiteRook_Opening[64];
char Score::WhiteRook_Endgame[64];
char Score::WhiteQueen_Opening[64];
char Score::WhiteQueen_Endgame[64];
char Score::WhiteKing_Opening[64];
char Score::WhiteKing_Endgame[64];

int Score::WhiteKnightBlocker[64];
int Score::WhiteKnightOutpost[64];

int Score::RookOnOpenFile[8]        = { 15, 15, 17, 20, 20, 17, 15, 15 };

static void getSymmetricSquareValueTableCh( char * dst, const char * src )
{
    for( int square=0; square<64; square++ ) {
        int file = FileOfSquare( square );
        int rank = RankOfSquare( square );

        int symmetric_square = FileRankToSquare( file, 7-rank );

        dst[symmetric_square] = src[square];
    }
}

void Score::initialize()
{
    int i;

    getSymmetricSquareValueTableCh( WhiteKnight_Opening, BlackKnight_Opening );
    getSymmetricSquareValueTableCh( WhiteKnight_Endgame, BlackKnight_Endgame );
    getSymmetricSquareValueTableCh( WhiteBishop_Opening, BlackBishop_Opening );
    getSymmetricSquareValueTableCh( WhiteBishop_Endgame, BlackBishop_Endgame );
    getSymmetricSquareValueTableCh( WhiteRook_Opening, BlackRook_Opening );
    getSymmetricSquareValueTableCh( WhiteRook_Endgame, BlackRook_Endgame );
    getSymmetricSquareValueTableCh( WhiteQueen_Opening, BlackQueen_Opening );
    getSymmetricSquareValueTableCh( WhiteQueen_Endgame, BlackQueen_Endgame );
    getSymmetricSquareValueTableCh( WhiteKing_Opening, BlackKing_Opening );
    getSymmetricSquareValueTableCh( WhiteKing_Endgame, BlackKing_Endgame );

    getSymmetricSquareValueTable( WhiteKnightBlocker, BlackKnightBlocker );
    getSymmetricSquareValueTable( WhiteKnightOutpost, BlackKnightOutpost );

    // Piece values: these tables must always be constructed dynamically,
    // because we must keep available the freedom of changing the internal
    // bit representation of a piece (i.e. which bits identify the piece
    // and the side... see the definitions of PieceSide and PieceType)
    for( i=0; i<16; i++ ) {
        int value = 0;

        // Basic idea for evaluation: no single piece can increment
        // the defect count alone, but Q+P or R+R or say B+B+N are
        // all worth one defect.

        switch( i ) {
        case ::BlackPawn:
        case ::WhitePawn:
            value = Score::Pawn;    
            break;
        case ::BlackKnight: 
        case ::WhiteKnight: 
            value = Score::Knight;  
            break;
        case ::BlackBishop: 
        case ::WhiteBishop: 
            value = Score::Bishop;  
            break;
        case ::BlackRook:   
        case ::WhiteRook:   
            value = Score::Rook;    
            break;
        case ::BlackQueen:  
        case ::WhiteQueen:  
            value = Score::Queen;
            break;
        case ::BlackKing:   
        case ::WhiteKing:   
            value = Score::King;    
            break;
        }

        PieceAbs[i] = value;
        Piece[i] = PieceSide(i) == Black ? -value : +value;
    }

    // Prepare some tables with pointers to the piece/square tables that might
    // be useful elsewhere
    for( i=0; i<16; i++ ) {
        ByPiece_Opening[i] = 0;
        ByPiece_Endgame[i] = 0;
    }

    ByPiece_Opening[ ::BlackKnight ]  = BlackKnight_Opening;
    ByPiece_Opening[ ::BlackBishop ]  = BlackBishop_Opening;
    ByPiece_Opening[ ::BlackRook ]  = BlackRook_Opening;
    ByPiece_Opening[ ::BlackQueen ]  = BlackQueen_Opening;
    ByPiece_Opening[ ::BlackKing ]  = BlackKing_Opening;
    ByPiece_Opening[ ::WhiteKnight ]  = WhiteKnight_Opening;
    ByPiece_Opening[ ::WhiteBishop ]  = WhiteBishop_Opening;
    ByPiece_Opening[ ::WhiteRook ]  = WhiteRook_Opening;
    ByPiece_Opening[ ::WhiteQueen ]  = WhiteQueen_Opening;
    ByPiece_Opening[ ::WhiteKing ]  = WhiteKing_Opening;

    ByPiece_Endgame[ ::BlackKnight ]  = BlackKnight_Endgame;
    ByPiece_Endgame[ ::BlackBishop ]  = BlackBishop_Endgame;
    ByPiece_Endgame[ ::BlackRook ]  = BlackRook_Endgame;
    ByPiece_Endgame[ ::BlackQueen ]  = BlackQueen_Endgame;
    ByPiece_Endgame[ ::BlackKing ]  = BlackKing_Endgame;
    ByPiece_Endgame[ ::WhiteKnight ]  = WhiteKnight_Endgame;
    ByPiece_Endgame[ ::WhiteBishop ]  = WhiteBishop_Endgame;
    ByPiece_Endgame[ ::WhiteRook ]  = WhiteRook_Endgame;
    ByPiece_Endgame[ ::WhiteQueen ]  = WhiteQueen_Endgame;
    ByPiece_Endgame[ ::WhiteKing ]  = WhiteKing_Endgame;
}

void Score::getSymmetricSquareValueTable( int * dst, const int * src )
{
    for( int square=0; square<64; square++ ) {
        int file = FileOfSquare( square );
        int rank = RankOfSquare( square );

        int symmetric_square = FileRankToSquare( file, 7-rank );

        dst[symmetric_square] = src[square];
    }
}
